@@ -1,20 +1,22 @@
+import pickle
 import threading
 import time
-
 from datetime import timedelta
-from functools import wraps, update_wrapper, partial
-from typing import Iterator, Generator
+from functools import partial, update_wrapper, wraps
+from queue import Empty, Full, PriorityQueue
+from threading import Lock
+from typing import Generator, Iterator
 
 from celery import _state, current_app, current_task, group
 from celery.exceptions import TimeoutError as CeleryTimeoutError
 from celery.result import AsyncResult, GroupResult, ResultSet
 from celery.utils.abstract import CallableSignature, CallableTask
 
-from .backends import default_backend
+from .backends import DEFAULT_BACKEND
 from .mixins import LoggerMixin
-from .states import PROGRESS
 from .signals import subtask_success
-from .utils import gen_chunk, total_size, is_uuid_format
+from .states import PROGRESS
+from .utils import gen_chunk, is_uuid_format, total_size
 
 
 class TaskThread(threading.Thread):
@@ -144,8 +146,9 @@ class dispatch(LoggerMixin, threading.local):
     def __init__(self, *args, **kwargs):
         # initialize configuration
         self._result_backend = current_app.conf.get('dispatcher_result_backend')
-        self._chunk_size = current_app.conf.get('dispatcher_chunk_size', 1000)
-        self._subtask_timeout = current_app.conf.get('dispatcher_subtask_timeout', 60*60)
+        self._batch_size = current_app.conf.get('dispatcher_batch_size', 1000)
+        self._poll_size = current_app.conf.get('dispatcher_poll_size', 1000)
+        self._subtask_timeout = current_app.conf.get('dispatcher_subtask_timeout', 60 * 60)
         self._result_expires = current_app.conf.get('result_expires')
         if isinstance(self._result_expires, timedelta):
             self._result_expires = self._result_expires.total_seconds()
@@ -153,8 +156,8 @@ class dispatch(LoggerMixin, threading.local):
         # initialize dispatching
         progress_update_frequency = current_app.conf.get('dispatcher_progress_update_frequency', 1)
         self._progress_manager = ProgressManager(progress_update_frequency)
-        self._dispatch_finished = threading.Event()
         self._backend = None  # set in _create_task_wrapper
+        self._auto_ignore = True
 
         if len(args) == 1 and callable(args[0]):
             self._wrapped_directly = True
@@ -165,7 +168,8 @@ class dispatch(LoggerMixin, threading.local):
             update_wrapper(self, task)
             origin_call = self.__call__
             self.__call__ = partial(origin_call, self)
-            self.__call__ = wraps(origin_call)(self.__call__)  # partial miss to copy some attributes from the original function
+            # partial miss to copy some attributes from the original function
+            self.__call__ = wraps(origin_call)(self.__call__)
         else:
             self._wrapped_directly = False
             self._wrapper = self._create_task_wrapper(*args, **kwargs)
@@ -179,7 +183,7 @@ class dispatch(LoggerMixin, threading.local):
             self._wrapped_task = self._wrapper(task)
             return self._wrapped_task
 
-    def _create_task_wrapper(self, options=None, receiver=None, backend=None):
+    def _create_task_wrapper(self, options=None, receiver=None, backend=None, auto_ignore=True):
         def wrapper(task):
             @wraps(task)
             def wrapped_task(*args, **kwargs):
@@ -188,9 +192,12 @@ class dispatch(LoggerMixin, threading.local):
                 if not isinstance(gen_task, Generator):
                     raise RuntimeError('Task should be generator: {0!r}'.format(task))
 
-                self._dispatch_from_task(gen_task, **options)
+                self._progress_manager.init_progress()
+                stash_finished = threading.Event()
+
+                self._dispatch_from_task(stash_finished, gen_task, **options)
                 # start collecting
-                self._collect_from_task(receiver, auto_ignore=True)
+                self._collect_from_task(stash_finished, receiver)
 
             return wrapped_task
 
@@ -200,7 +207,9 @@ class dispatch(LoggerMixin, threading.local):
         if backend:
             self._backend = backend(self._result_backend)
         else:
-            self._backend = default_backend(self._result_backend)
+            self._backend = DEFAULT_BACKEND(self._result_backend)
+
+        self._auto_ignore = auto_ignore
 
         return wrapper
 
@@ -209,7 +218,7 @@ class dispatch(LoggerMixin, threading.local):
         root_id = current_task.request.root_id
         return 'celery-task-meta-{0}:subtask'.format(root_id)
 
-    def _dispatch_from_task(self, *tasks, **options):
+    def _dispatch_from_task(self, stash_finished, *tasks, **options):
         if len(tasks) == 1:
             tasks = tasks[0]
 
@@ -225,102 +234,139 @@ class dispatch(LoggerMixin, threading.local):
         # `group.tasks` and `GroupResult`
         #
         # results = group(task for task in tasks)(**options)
-        # self._handle_results(results, self.receive_result)
+        # self._handle_result(results, self.receive_result)
 
-        self._progress_manager.init_progress()
-        self._dispatch_finished.clear()
         self.logger.info('Start dispatching')
 
-        dispatch_thread = TaskThread(
-            target=self._store_sub_task_results,
-            args=(self._progress_manager, self._dispatch_finished, tasks,),
+        stash_thread = TaskThread(
+            target=self._stash_subtask_results,
+            args=(self._progress_manager, stash_finished, tasks,),
             kwargs=options)
-        dispatch_thread.start()
-        self.logger.debug('Dispatching sub-tasks: %r, options: %r', tasks, options)
+        stash_thread.start()
+        self.logger.debug('Dispatching subtasks: %r, options: %r', tasks, options)
 
-    def _store_sub_task_results(self, progress_manager, dispatch_finished, tasks, **options):
-        for results in gen_chunk(self._apply_tasks(tasks, **options), self._chunk_size):
+    def _stash_subtask_results(self, progress_manager, stash_finished, tasks, **options):
+        for results in gen_chunk(self._apply_tasks(tasks, **options), self._batch_size):
             self.logger.debug('%s tasks have been applied', len(results))
             progress_manager.update_progress_total(len(results))
-            self._backend.bulk_push(self._dispatch_key, map(lambda r: r.task_id, results), self._result_expires)
+            self._backend.bulk_push(self._dispatch_key, map(pickle.dumps, results), self._result_expires)
 
         time.sleep(0)
-        dispatch_finished.set()
+        stash_finished.set()
         self.logger.info('Finished dispatching')
 
-    def _apply_tasks(self, tasks, **options) -> Iterator[AsyncResult]:
+    def _apply_tasks(self, tasks, **common_options) -> Iterator[AsyncResult]:
         with current_app.producer_or_acquire() as producer:
-            for yield_task in tasks:
-                task, args, kwargs, task_options = self._parse_yield_task(yield_task)
-                _options = options.copy()
-                _options.update(task_options)
-                result = task.apply_async(args=args, kwargs=kwargs, producer=producer, add_to_parent=False, **_options)
-                task_id = result.task_id
-                self.logger.debug('Task %s applied', task_id)
+            for task_info in tasks:
+                task, args, kwargs, task_options = self._parse_task_info(task_info)
+                options = common_options.copy()
+                options.update(task_options)
+                result = task.apply_async(args=args, kwargs=kwargs, producer=producer, add_to_parent=False, **options)
+                self.logger.debug('Task %s applied', result.task_id)
                 yield result
 
-    def _parse_yield_task(self, yield_task):
+    def _parse_task_info(self, task_info):
         task = None
         args = ()
         kwargs = {}
         options = {}
-        if isinstance(yield_task, CallableTask):
-            task = yield_task
-        elif isinstance(yield_task, (tuple, list)):
-            if len(yield_task) == 1:
-                task = yield_task[0]
-            elif len(yield_task) == 2:
-                task, args = yield_task
-            elif len(yield_task) == 3:
-                task, args, kwargs = yield_task
-            elif len(yield_task) == 4:
-                task, args, kwargs, options = yield_task
+        if isinstance(task_info, CallableTask):
+            task = task_info
+        elif isinstance(task_info, (tuple, list)):
+            if len(task_info) == 1:
+                task = task_info[0]
+            elif len(task_info) == 2:
+                task, args = task_info
+            elif len(task_info) == 3:
+                task, args, kwargs = task_info
+            elif len(task_info) == 4:
+                task, args, kwargs, options = task_info
 
         if not task:
-            raise ValueError('Invalid yield task: {0!r}'.format(yield_task))
+            raise ValueError('Invalid task info: {0!r}'.format(task_info))
 
         return task, args, kwargs, options
 
-    def _collect_from_task(self, receiver, auto_ignore: bool = True):
+    def _collect_from_task(self, stash_finished, receiver):
         self.logger.info('Start collecting')
-        for result in self._gen_sub_result():
-            try:
-                self._handle_results(result, receiver, auto_ignore=auto_ignore)
-            except CeleryTimeoutError:
-                self.logger.exception('Results timeout expired, task_id: %s', result.task_id)
+
+        restore_finished = threading.Event()
+        handle_lock = Lock()
+        result_queue = PriorityQueue(self._poll_size)
+
+        restore_thread = TaskThread(
+            target=self._restore_subtask_results,
+            args=(stash_finished, restore_finished, handle_lock, result_queue))
+        restore_thread.start()
+
+        while True:
+            if result_queue.empty() and restore_finished.is_set():
+                break
+
+            with handle_lock:
+                try:
+                    priority, result = result_queue.get(False)
+                except Empty:
+                    continue
+
+                timeout = self._subtask_timeout if restore_finished.is_set() else 1
+                try:
+                    self._handle_result(result, receiver, timeout=timeout)
+                except CeleryTimeoutError:
+                    if restore_finished.is_set():
+                        self.logger.debug('Subtask timeout, task_id: %s', result.task_id)
+                        self._progress_manager.update_progress_completed()
+                    else:
+                        result_queue.put((priority + 10, result))
+                except Exception:
+                    self.logger.debug('Subtask raised exception, task_id: %s', result.task_id)
+                    self._progress_manager.update_progress_completed()
 
         self.logger.info('Finished collecting')
 
-    def _gen_sub_result(self, interval=0.5) -> Iterator[AsyncResult]:
+    def _restore_subtask_results(self, stash_finished, restore_finished, handle_lock, result_queue, interval=0.5):
         while True:
-            task_id = self._backend.pop(self._dispatch_key)
-            if not task_id:
-                if self._dispatch_finished.is_set():
-                    return
+            result = self._backend.pop(self._dispatch_key)
+            if not result:
+                if stash_finished.is_set():
+                    break
+
                 time.sleep(interval)
                 continue
-            result = current_app.AsyncResult(task_id.decode('utf-8'))
-            yield result
 
-    def _handle_results(self, results, receiver, other_worker=False, auto_ignore=False, timeout=None, interval=0.5,
-                        on_interval=None, propagate=False, disable_sync_subtasks=False):
-        callback = partial(self._on_task_finished, receiver=receiver, auto_ignore=auto_ignore)
-        timeout = timeout or self._subtask_timeout
-        self.logger.debug('Waiting for the results, task_id: %s, timeout: %s', results, timeout)
+            result = pickle.loads(result)
+
+            while True:
+                try:
+                    with handle_lock:
+                        result_queue.put((time.time(), result), False)
+                except Full:
+                    time.sleep(interval)
+                    continue
+                else:
+                    break
+
+        with handle_lock:
+            restore_finished.set()
+
+    def _handle_result(self, result, receiver, other_worker=False, timeout=None, interval=0.5, on_interval=None,
+                       propagate=True, disable_sync_subtasks=False):
+        callback = partial(self._on_task_finished, receiver=receiver)
+        self.logger.debug('Waiting for the result, task_id: %s, timeout: %s', result, timeout)
         if other_worker:
-            self.logger.debug('Task is not ready on other worker, task_id: %s', results.task_id)
-            self._wait_until_ready(results, timeout)
+            self.logger.debug('Task is not ready on other worker, task_id: %s', result.task_id)
+            self._wait_until_ready(result, timeout)
 
-        if isinstance(results, AsyncResult):
-            result = results.get(
+        if isinstance(result, AsyncResult):
+            value = result.get(
                 timeout=timeout,
                 interval=interval,
                 on_interval=on_interval,
                 propagate=propagate,
                 disable_sync_subtasks=disable_sync_subtasks)
-            callback(results.task_id, result)
-        elif isinstance(results, ResultSet):
-            results.get(
+            callback(result.task_id, value)
+        elif isinstance(result, ResultSet):
+            result.get(
                 callback=callback,
                 timeout=timeout,
                 interval=interval,
@@ -328,10 +374,10 @@ class dispatch(LoggerMixin, threading.local):
                 propagate=propagate,
                 disable_sync_subtasks=disable_sync_subtasks)
         else:
-            raise ValueError('Invalid results type: {0!r}'.format(results))
+            raise ValueError('Invalid result type: {0!r}'.format(result))
 
-        if auto_ignore:
-            results.forget()
+        if self._auto_ignore:
+            result.forget()
 
     def _wait_until_ready(self, result, timeout=None, interval=0.5):
         time_start = time.monotonic()
@@ -341,21 +387,21 @@ class dispatch(LoggerMixin, threading.local):
                 raise TimeoutError('Failed to wait for the result, task_id: %s', result.task_id)
             time.sleep(interval)
 
-    def _on_task_finished(self, task_id, value, receiver, auto_ignore=False):
+    def _on_task_finished(self, task_id, value, receiver):
         if isinstance(value, Exception):
             self.logger.error('Failed to save result, task_id: %s', task_id, exc_info=value)
 
         elif isinstance(value, GroupResult):
             self.logger.debug('Received GroupResult, task_id: %s', task_id)
             self._progress_manager.update_progress_total(len(value))
-            self._handle_results(value, receiver=receiver, auto_ignore=auto_ignore)
+            self._handle_result(value, receiver=receiver)
 
         elif self._is_groupresult_meta(value):
             self.logger.debug('Received GroupResult META, task_id: %s', task_id)
-            for results in gen_chunk(self._gen_result_from_groupresult_meta(value), self._chunk_size):
+            for results in gen_chunk(self._gen_result_from_groupresult_meta(value), self._batch_size):
                 self._progress_manager.update_progress_total(len(results))
                 for result in results:
-                    self._handle_results(result, receiver=receiver, other_worker=True, auto_ignore=auto_ignore)
+                    self._handle_result(result, receiver=receiver, other_worker=True)
 
         else:
             self.logger.debug('Received common result, task_id: %s, size: %d bytes', task_id, total_size(value))
